#[macro_use]
extern crate nom;
use std::fmt;
use std::str::FromStr;
mod display;
use nom::types::CompleteStr as ComStr;
pub use nom::Err;
pub use nom::ErrorKind;

#[derive(Clone, Eq, PartialEq, Debug)]
pub struct ParseError(pub ErrorKind);

impl std::error::Error for ParseError {}
impl fmt::Display for ParseError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "cfg parse error: {:?}", self.0)
    }
}

#[derive(Debug, Clone, Eq, PartialEq)]
pub enum Cfg {
    Any(Vec<Cfg>),
    All(Vec<Cfg>),
    Not(Box<Cfg>),
    Equal(String, String),
    Is(String),
}

#[derive(Debug, Clone, Eq, PartialEq)]
pub enum Target {
    Triple {
        arch: String,
        vendor: String,
        os: String,
        env: Option<String>,
    },
    Cfg(Cfg),
}

/// Parse expression `cfg(…)`
pub fn parse_cfg(cfg_str: &str) -> Result<Cfg, nom::Err<ComStr<'_>>> {
    cfg(ComStr(cfg_str)).map(|(_, res)| res)
}

/// Parse expression valid as Cargo target (`cfg()` or `target-triple`)
pub fn parse_target(target_str: &str) -> Result<Target, nom::Err<ComStr<'_>>> {
    target(ComStr(target_str)).map(|(_, res)| res)
}

named!(cfg<ComStr<'_>, Cfg>,
  preceded!(
        ws!(tag!("cfg")),
        delimited!(
            ws!(char!('(')),
            expr,
            terminated!(
                ws!(char!(')')),
                eof!()))));

named!(expr<ComStr<'_>, Cfg>,
  ws!(alt!(any | all | not | equal | is)));

named!(any<ComStr<'_>, Cfg>,
    map!(
        preceded!(
            tag!("any"),
            expr_list),
        |e| Cfg::Any(e)));

named!(all<ComStr<'_>, Cfg>,
    map!(
        preceded!(
            tag!("all"),
            expr_list),
        |e| Cfg::All(e)));

named!(not<ComStr<'_>, Cfg>,
    map!(
        preceded!(
            tag!("not"),
            delimited!(
                ws!(char!('(')),
                expr,
                ws!(char!(')')))),
        |e| Cfg::Not(Box::new(e))));

named!(expr_list<ComStr<'_>, Vec<Cfg>>,
    delimited!(
        ws!(char!('(')),
        separated_list!(
            char!(','),
            expr),
        ws!(char!(')'))));

named!(equal<ComStr<'_>, Cfg>,
    map!(
        separated_pair!(
            bareword,
            ws!(char!('=')),
            alt!(literal | bareword)),
        |(a,b)| Cfg::Equal(a,b)));

named!(bareword<ComStr<'_>, String>,
    map!(take_while1!(|c: char| c.is_alphanumeric() || c == '_' || c == '-'), |r| r.to_string()));

named!(is<ComStr<'_>, Cfg>,
    map!(bareword, |s| Cfg::Is(s)));

named!(literal<ComStr<'_>, String>,
    map!(
        delimited!(
            char!('"'),
            escaped!(is_not!("\"\\"), '\\', one_of!("\"n\\")), // FIXME: needs escaped_transform
            char!('"')),
        |s| s.to_string()));

named!(target<ComStr<'_>, Target>,
    alt!(
        terminated!(ws!(triple), eof!()) |
        map!(
            cfg,
            |c| Target::Cfg(c))));

named!(alnum<ComStr<'_>, String>,
    map!(take_while1!(|c: char| c.is_alphanumeric() || c == '_'), |r| r.to_string()));

named!(triple<ComStr<'_>, Target>,
    do_parse!(
        arch: alnum >>
        char!('-') >>
        vendor: alnum >>
        char!('-') >>
        os: alnum >>
        env: opt!(preceded!(
            char!('-'),
            alnum)) >>
        (Target::Triple {arch, vendor, os, env})));

impl FromStr for Cfg {
    type Err = ParseError;
    fn from_str(cfg_str: &str) -> Result<Cfg, Self::Err> {
        parse_cfg(cfg_str).map_err(|e| ParseError(e.into_error_kind()))
    }
}

impl FromStr for Target {
    type Err = ParseError;
    fn from_str(cfg_str: &str) -> Result<Target, Self::Err> {
        parse_target(cfg_str).map_err(|e| ParseError(e.into_error_kind()))
    }
}

#[test]
fn parses() {
    assert_eq!(Ok(Cfg::Is("a".to_string())), parse_cfg(r#"cfg(a)"#));
    assert_eq!(Ok(Target::Cfg(Cfg::Is("a".to_string()))), parse_target(r#"cfg(a)"#));
    assert_eq!(Ok(Target::Triple{
        arch: "sparcv9".to_string(),
        vendor: "sun".to_string(),
        os: "solaris".to_string(),
        env: None,
    }), parse_target("sparcv9-sun-solaris"));
    assert_eq!(Ok(Target::Triple {
        arch: "armv5te".to_string(),
        vendor: "unknown".to_string(),
        os: "linux".to_string(),
        env: Some("musleabi".to_string()),
    }), "  armv5te-unknown-linux-musleabi  ".parse());
    assert_eq!(Ok(Cfg::Any(vec![Cfg::Is("ha".to_string())])), r#"cfg(any(ha))"#.parse());
    assert_eq!(Ok(Cfg::All(vec![Cfg::Is("ha".to_string())])), parse_cfg(r#"cfg(all(ha))"#));
    assert_eq!(Ok(Cfg::Not(Box::new(Cfg::Is("ha".to_string())))), parse_cfg(r#"cfg(not(ha))"#));
    assert_eq!(Ok(Cfg::Any(vec![
        Cfg::Is("a".to_string()), Cfg::Is("b".to_string())
        ])), parse_cfg(r#"cfg( any ( a, b))"#));
    assert_eq!(Ok(Cfg::Is("a".to_string())), r#"cfg ( a )"#.parse());
    assert_eq!(Ok(Cfg::Equal("a".to_string(),"b".to_string())), parse_cfg(r#" cfg(a=b) "#));
    assert_eq!(Ok(Cfg::Equal("a".to_string(),"b".to_string())), parse_cfg(r#" cfg(  a  =  b  ) "#));
    assert_eq!(Ok(Cfg::Any(vec![Cfg::Equal("a".to_string(),"b".to_string())])), parse_cfg(r#"cfg(any(a=b))"#));
    assert_eq!(Ok(Cfg::All(vec![Cfg::Equal("a".to_string(),"b".to_string())])), parse_cfg(r#"cfg(all(a=b))"#));
    assert_eq!(Ok(Cfg::Not(Box::new(Cfg::Equal("a".to_string(),"b".to_string())))), parse_cfg(r#"cfg(not(a=b))"#));
    assert_eq!(Ok(Cfg::Not(Box::new(Cfg::Equal("a_b".to_string(),"b".to_string())))), parse_cfg(r#" cfg( not( a_b  =  b ) ) "#));
    assert_eq!(Ok(Cfg::Not(Box::new(Cfg::Equal("a".to_string(),"b".to_string())))), parse_cfg(r#" cfg( not( a  =  "b" ) ) "#));
    assert!(parse_cfg(r#" cfg( not( a  =  "b\"\\" ) ) "#).is_ok());
}

#[test]
fn targets() {
    for t in ["aarch64-apple-ios", "aarch64-linux-android", "aarch64-unknown-fuchsia",
    "aarch64-unknown-linux-gnu", "aarch64-unknown-linux-musl", "arm-linux-androideabi",
    "arm-unknown-linux-gnueabi", "arm-unknown-linux-gnueabihf", "arm-unknown-linux-musleabi",
    "arm-unknown-linux-musleabihf", "armv5te-unknown-linux-gnueabi",
    "armv5te-unknown-linux-musleabi", "armv7-apple-ios", "armv7-linux-androideabi",
    "armv7-unknown-linux-gnueabihf", "armv7-unknown-linux-musleabihf", "armv7s-apple-ios",
    "asmjs-unknown-emscripten", "i386-apple-ios", "i586-pc-windows-msvc",
    "i586-unknown-linux-gnu", "i586-unknown-linux-musl", "i686-apple-darwin",
    "i686-linux-android", "i686-pc-windows-gnu", "i686-pc-windows-msvc",
    "i686-unknown-freebsd", "i686-unknown-linux-gnu", "i686-unknown-linux-musl",
    "mips-unknown-linux-gnu", "mips-unknown-linux-musl", "mips64-unknown-linux-gnuabi64",
    "mips64el-unknown-linux-gnuabi64", "mipsel-unknown-linux-gnu", "mipsel-unknown-linux-musl",
    "powerpc-unknown-linux-gnu", "powerpc64-unknown-linux-gnu", "powerpc64le-unknown-linux-gnu",
    "s390x-unknown-linux-gnu", "sparc64-unknown-linux-gnu", "sparcv9-sun-solaris",
    "thumbv6m-none-eabi", "thumbv7em-none-eabi", "thumbv7em-none-eabihf",
    "thumbv7m-none-eabi", "wasm32-unknown-emscripten", "wasm32-unknown-unknown",
    "x86_64-apple-darwin", "x86_64-apple-ios", "x86_64-linux-android",
    "x86_64-pc-windows-gnu", "x86_64-pc-windows-msvc", "x86_64-rumprun-netbsd",
    "x86_64-sun-solaris", "x86_64-unknown-cloudabi", "x86_64-unknown-freebsd",
    "x86_64-unknown-fuchsia", "x86_64-unknown-linux-gnu", "x86_64-unknown-linux-gnux32",
    "x86_64-unknown-linux-musl", "x86_64-unknown-netbsd", "x86_64-unknown-redox"].iter() {
        assert!(parse_target(t).is_ok(), t);
    }
}

#[test]
fn garbage() {
    assert!(parse_target("x86_64-pc-windows-gnu!(#$@#)").is_err());
    assert!(parse_target("cfg(ok)--not ok").is_err());
}
